"""
pytest -v tests/test_splitting.py
"""
import logging
import pytest
from wayfarer import splitter, functions, loader, routing
from tests.helper import simple_features


@pytest.mark.parametrize("use_reverse_lookup", [(True), (False)])
def test_split_network_edge(use_reverse_lookup):

    feats = simple_features()
    net = loader.load_network_from_geometries(
        feats, use_reverse_lookup=use_reverse_lookup
    )

    assert len(net.nodes()) == 4
    assert len(net.edges()) == 3

    # split the second edge
    edge_id_to_split = 2
    measure = 50
    splitter.split_network_edge(net, edge_id_to_split, [measure])

    split_node_id = splitter.create_split_key(edge_id_to_split, measure)
    assert split_node_id in net.nodes(), "The split node has been added to the network"

    # after the split operation there should be one extra edge
    # and one extra node
    assert len(net.nodes()) == 5
    assert len(net.edges()) == 4

    # the split edges no longer have the EdgeId as a key
    split_edges = functions.get_edge_by_attribute(net, "EDGE_ID", edge_id_to_split)
    split_edges = list(split_edges)
    assert len(split_edges) == 2

    # if use_reverse_lookup: print(net.graph["keys"].keys())

    split_edge_id = splitter.create_split_key(edge_id_to_split, measure)

    split_edge = functions.get_edge_by_key(net, split_edge_id, with_data=True)
    assert split_edge.attributes["LEN_"] == 50
    assert split_edge.attributes["OFFSET"] == 0

    split_edge_id = splitter.create_split_key(edge_id_to_split, 100)
    split_edge = functions.get_edge_by_key(net, split_edge_id, with_data=True)
    assert split_edge.attributes["LEN_"] == 50
    assert split_edge.attributes["OFFSET"] == 50


@pytest.mark.parametrize("use_reverse_lookup", [(True), (False)])
def test_multiple_split_network_edge(use_reverse_lookup):

    feats = simple_features()
    net = loader.load_network_from_geometries(
        feats, use_reverse_lookup=use_reverse_lookup
    )

    # split the second edge multiple times
    edge_id_to_split = 2
    splitter.split_network_edge(net, edge_id_to_split, [20, 40, 60])

    if use_reverse_lookup:
        print(net.graph["keys"].keys())

    # after the split operation there should be one extra edge
    # and one extra node
    assert len(net.nodes()) == 7
    assert len(net.edges()) == 6


@pytest.mark.parametrize("use_reverse_lookup", [(True), (False)])
def test_double_split_network_edge(use_reverse_lookup):
    feats = simple_features()
    net = loader.load_network_from_geometries(
        feats, use_reverse_lookup=use_reverse_lookup
    )

    # split the second edge multiple times
    edge_id_to_split = 2
    splitter.split_network_edge(net, edge_id_to_split, [20, 40, 60])

    split_edge_id = splitter.create_split_key(edge_id_to_split, 40)

    if use_reverse_lookup:
        print(net.graph["keys"].keys())

    splitter.split_network_edge(net, split_edge_id, [10])

    if use_reverse_lookup:
        print(net.graph["keys"].keys())

    assert len(net.nodes()) == 8
    assert len(net.edges()) == 7


@pytest.mark.parametrize("use_reverse_lookup", [(True), (False)])
def test_split_invalid_measure(use_reverse_lookup):

    feats = simple_features()
    net = loader.load_network_from_geometries(
        feats, use_reverse_lookup=use_reverse_lookup
    )
    error = False

    try:
        # 150 is greater than edge length
        splitter.split_network_edge(net, 2, [150])
    except ValueError:
        error = True

    assert error


@pytest.mark.parametrize("use_reverse_lookup", [(True), (False)])
def test_split_invalid_measure2(use_reverse_lookup):

    feats = simple_features()
    net = loader.load_network_from_geometries(
        feats, use_reverse_lookup=use_reverse_lookup
    )
    error = False

    try:
        # 100 is equal to edge length
        splitter.split_network_edge(net, 2, [100])
    except ValueError:
        error = True

    assert error


@pytest.mark.parametrize("use_reverse_lookup", [(True), (False)])
def test_split_with_points(use_reverse_lookup):

    feats = simple_features()
    net = loader.load_network_from_geometries(
        feats, use_reverse_lookup=use_reverse_lookup
    )

    recs = [
        {"EDGE_ID": 1, "POINT_ID": "A", "MEASURE": 50, "LEN_": 12},
        {"EDGE_ID": 2, "POINT_ID": "B", "MEASURE": 20, "LEN_": 14},
        {"EDGE_ID": 2, "POINT_ID": "C", "MEASURE": 40, "LEN_": 16},
    ]

    splitter.split_with_points(net, recs)

    # run a test solve to the newly added point
    point_id = "B"
    edges = routing.solve_shortest_path(net, "0|0", point_id, with_direction_flag=True)
    # print(edges)
    assert len(edges) == 4

    # run a solve from start to finish across the original (now split) edges
    edges = routing.solve_shortest_path(net, "0|0", "300|0", with_direction_flag=True)
    # print(edges)
    assert len(edges) == 6


def test_doctest():
    import doctest

    print(doctest.testmod(splitter))


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    # test_multiple_split_network_edge(True)
    # test_split_invalid_measure(True)
    # test_split_invalid_measure2(True)
    # test_multiple_split_network_edge(True)
    # test_multiple_split_network_edge(True)
    # test_double_split_network_edge(True)
    test_split_with_points(True)
    test_split_network_edge(True)
    print("Done!")



"""
pytest -v tests/test_functions.py
"""

import logging
import pytest
import wayfarer
from wayfarer import loader, functions, Edge, WITH_DIRECTION_FIELD
import networkx


@pytest.mark.parametrize("use_reverse_lookup", [(True), (False)])
def test_get_edge_by_key(use_reverse_lookup):
    recs = [{"EDGE_ID": 1, "LEN_": 100, "NODEID_FROM": 0, "NODEID_TO": 100}]
    net = loader.load_network_from_records(recs, use_reverse_lookup=use_reverse_lookup)

    # print(net.graph.keys())
    if use_reverse_lookup:
        assert "keys" in net.graph.keys()  # dict_keys(['name', 'keys'])

    edge = functions.get_edge_by_key(net, 1, with_data=False)

    assert edge == (0, 100, 1, {})

    edge = functions.get_edge_by_key(net, 1, with_data=True)

    assert edge == (
        0,
        100,
        1,
        {"NODEID_FROM": 0, "LEN_": 100, "EDGE_ID": 1, "NODEID_TO": 100},
    )


def test_get_edge_by_key_missing():
    recs = [{"EDGE_ID": 1, "LEN_": 100, "NODEID_FROM": 0, "NODEID_TO": 100}]
    net = loader.load_network_from_records(recs)

    with pytest.raises(KeyError):
        functions.get_edge_by_key(net, -1)


def test_to_edge():
    edge = wayfarer.to_edge((0, 1, 1, {"LEN_": 10}))
    assert edge == Edge(start_node=0, end_node=1, key=1, attributes={"LEN_": 10})


def test_to_edge_no_attributes():
    """
    If no attributes are supplied in a tuple then an empty dict is returned
    """

    edge = wayfarer.to_edge((0, 1, 1))
    assert edge == Edge(start_node=0, end_node=1, key=1, attributes={})


def test_get_multiple_edges_by_attribute():
    net = networkx.MultiGraph()

    net.add_edge(0, 0, key=1, **{"EDGE_ID": 1, "LEN_": 100})
    net.add_edge(0, 0, key=2, **{"EDGE_ID": 1, "LEN_": 100})

    edges = list(functions.get_edge_by_attribute(net, "EDGE_ID", 1))
    assert edges == [
        (0, 0, 1, {"EDGE_ID": 1, "LEN_": 100}),
        (0, 0, 2, {"EDGE_ID": 1, "LEN_": 100}),
    ]


def test_get_single_path_from_nodes():
    net = networkx.MultiGraph()

    net.add_edge(0, 1, key=1, **{"EDGE_ID": 1, "LEN_": 100})

    paths = functions.get_all_paths_from_nodes(net, [0, 1])
    key_list = [[e.key for e in p] for p in paths]
    assert key_list == [[1]]


def test_get_all_paths_from_nodes():
    net = networkx.MultiGraph()

    net.add_edge(0, 1, key=1, **{"EDGE_ID": 1, "LEN_": 100})
    net.add_edge(1, 2, key=2, **{"EDGE_ID": 2, "LEN_": 100})
    net.add_edge(2, 3, key=3, **{"EDGE_ID": 3, "LEN_": 100})

    paths = functions.get_all_paths_from_nodes(net, [0, 1, 2, 3])
    key_list = [[e.key for e in p] for p in paths]
    assert key_list == [[1, 2, 3]]


def test_get_all_paths_from_nodes_with_direction():
    net = networkx.MultiGraph()

    net.add_edge(
        0, 1, key=1, **{"EDGE_ID": 1, "LEN_": 100, "NODEID_FROM": 0, "NODEID_TO": 1}
    )
    # add a reversed edge in the middle of the path
    net.add_edge(
        1, 2, key=2, **{"EDGE_ID": 2, "LEN_": 100, "NODEID_FROM": 2, "NODEID_TO": 1}
    )
    net.add_edge(
        2, 3, key=3, **{"EDGE_ID": 3, "LEN_": 100, "NODEID_FROM": 2, "NODEID_TO": 3}
    )

    paths = list(
        functions.get_all_paths_from_nodes(net, [0, 1, 2, 3], with_direction_flag=True)
    )
    path = paths[0]

    key_list = [e.key for e in path]
    assert key_list == [1, 2, 3]
    assert path[0].attributes["WITH_DIRECTION"] is True
    assert path[1].attributes["WITH_DIRECTION"] is False
    assert path[2].attributes["WITH_DIRECTION"] is True


def test_get_all_complex_paths():
    """
    D- shape
    """
    net = networkx.MultiGraph()

    net.add_edge(0, 1, key="A")
    net.add_edge(1, 1, key="B")  # self-loop

    node_list = [0, 1]
    paths = functions.get_all_complex_paths(net, node_list)
    assert paths == [[0, 1], [0, 1, 1]]


def test_get_all_complex_paths2():
    """
    D-D shape
    """
    net = networkx.MultiGraph()

    net.add_edge(0, 0, key="A")
    net.add_edge(0, 1, key="B")
    net.add_edge(1, 1, key="C")

    node_list = [0, 1]
    paths = functions.get_all_complex_paths(net, node_list)
    assert paths == [[0, 1], [0, 0, 1], [0, 1, 1], [0, 0, 1, 1]]


def test_get_all_complex_paths3():
    """
    Self-loop in middle
    """
    net = networkx.MultiGraph()

    net.add_edge(0, 1, key="A")
    net.add_edge(1, 2, key="B")
    net.add_edge(2, 2, key="C")
    net.add_edge(2, 3, key="D")

    node_list = [0, 1, 2, 3]
    paths = functions.get_all_complex_paths(net, node_list)
    assert paths == [[0, 1, 2, 3], [0, 1, 2, 2, 3]]


def test_get_all_paths_from_nodes_loop():
    net = networkx.MultiGraph()

    net.add_edge(0, 1, key="A", **{"EDGE_ID": "A", "LEN_": 100})
    net.add_edge(1, 1, key="B", **{"EDGE_ID": "B", "LEN_": 100})  # self-loop

    paths = functions.get_all_paths_from_nodes(net, [0, 1])
    key_list = [[e.key for e in p] for p in paths]
    assert key_list == [["A"], ["A", "B"]]


def test_get_all_paths_from_nodes_loop2():
    net = networkx.MultiGraph()

    net.add_edge(0, 0, key="A", **{"EDGE_ID": "A", "LEN_": 100})
    net.add_edge(0, 1, key="B", **{"EDGE_ID": "B", "LEN_": 100})
    net.add_edge(1, 1, key="C", **{"EDGE_ID": "C", "LEN_": 100})

    paths = functions.get_all_paths_from_nodes(net, [0, 1])
    key_list = [[e.key for e in p] for p in paths]
    print(key_list)
    assert key_list == [["B"], ["A", "B"], ["B", "C"], ["A", "B", "C"]]


def test_get_all_paths_from_nodes_loop3():
    net = networkx.MultiGraph()

    net.add_edge(0, 1, key="A", **{"EDGE_ID": "A", "LEN_": 100})
    net.add_edge(1, 2, key="B", **{"EDGE_ID": "B", "LEN_": 100})
    net.add_edge(2, 2, key="C", **{"EDGE_ID": "C", "LEN_": 100})
    net.add_edge(2, 3, key="D", **{"EDGE_ID": "D", "LEN_": 100})

    paths = functions.get_all_paths_from_nodes(net, [0, 1, 2, 3])
    key_list = [[e.key for e in p] for p in paths]
    print(key_list)
    assert key_list == [["A", "B", "D"], ["A", "B", "C", "D"]]


def test_get_all_paths_from_nodes_loop4():
    # with parallel edge

    net = networkx.MultiGraph()

    net.add_edge(0, 1, key="A", **{"EDGE_ID": "A", "LEN_": 100})
    net.add_edge(1, 2, key="B", **{"EDGE_ID": "B", "LEN_": 100})
    net.add_edge(2, 1, key="B2", **{"EDGE_ID": "B2", "LEN_": 100})  # parallel edge
    net.add_edge(2, 2, key="C", **{"EDGE_ID": "C", "LEN_": 100})  # self-loop
    net.add_edge(2, 3, key="D", **{"EDGE_ID": "D", "LEN_": 100})

    paths = functions.get_all_paths_from_nodes(net, [0, 1, 2, 3])
    key_list = [[e.key for e in p] for p in paths]

    assert key_list == [
        ["A", "B", "D"],
        ["A", "B2", "D"],
        ["A", "B", "C", "D"],
        ["A", "B2", "C", "D"],
    ]


def test_get_all_paths_from_nodes_roundabout():
    """
    Get both the full and broken paths of a roundabout
    """
    net = networkx.MultiGraph()

    net.add_edge(0, 1, key="A", **{"EDGE_ID": "A", "LEN_": 100})
    net.add_edge(1, 2, key="B", **{"EDGE_ID": "B", "LEN_": 100})
    net.add_edge(2, 3, key="C", **{"EDGE_ID": "C", "LEN_": 100})
    net.add_edge(3, 0, key="D", **{"EDGE_ID": "D", "LEN_": 100})

    paths = functions.get_all_paths_from_nodes(net, [0, 1, 2, 3])
    key_list = [[e.key for e in p] for p in paths]
    assert key_list == [["A", "B", "C"], ["A", "B", "C", "D"]]


def test_get_all_paths_from_roundabout_with_loops():
    net = networkx.MultiGraph()

    net.add_edge(0, 0, key="AA", **{"EDGE_ID": "AA", "LEN_": 100})
    net.add_edge(0, 1, key="A", **{"EDGE_ID": "A", "LEN_": 100})
    net.add_edge(1, 2, key="B", **{"EDGE_ID": "B", "LEN_": 100})
    net.add_edge(2, 3, key="C", **{"EDGE_ID": "C", "LEN_": 100})
    net.add_edge(3, 0, key="D", **{"EDGE_ID": "D", "LEN_": 100})

    paths = functions.get_all_paths_from_nodes(net, [0, 1, 2, 3])
    key_list = [[e.key for e in p] for p in paths]
    assert key_list == [
        ["A", "B", "C"],
        ["AA", "A", "B", "C"],
        ["A", "B", "C", "D"],
        ["AA", "A", "B", "C", "D"],
    ]


def test_get_unattached_node():
    net = networkx.MultiGraph()
    edge1 = (0, 1)
    edge2 = (1, 2)

    net.add_edge(*edge1)
    net.add_edge(*edge2)

    n = functions.get_unattached_node(net, edge1)
    assert n == 0

    n = functions.get_unattached_node(net, edge2)
    assert n == 2


def test_get_unattached_node2():
    net = networkx.MultiGraph()
    edge1 = (0, 1)
    edge2 = (1, 2)
    edge3 = (2, 3)

    net.add_edge(*edge1)
    net.add_edge(*edge2)
    net.add_edge(*edge3)

    n = functions.get_unattached_node(net, edge1)
    assert n == 0

    n = functions.get_unattached_node(net, edge2)
    assert n is None

    n = functions.get_unattached_node(net, edge3)
    assert n == 3


def test_get_multiconnected_nodes():
    edges = [Edge(0, 1, "A", {}), Edge(1, 2, "B", {})]
    nodes = functions.get_multiconnected_nodes(edges, connections=2)
    assert nodes == [1]


def test_get_ordered_end_nodes():
    edges = [Edge(1, 2, "A", {}), Edge(0, 1, "B", {})]
    nodes = [0, 1]
    end_nodes = functions.get_ordered_end_nodes(edges, nodes)
    assert end_nodes == [1, 0]


def test_get_edges_from_nodes():
    net = networkx.MultiGraph()

    net.add_edge(0, 1, key="A", **{"EDGE_ID": "A", "LEN_": 100})
    net.add_edge(
        1, 2, key="B", **{"EDGE_ID": "B", "LEN_": 100, "NODEID_FROM": 1, "NODEID_TO": 2}
    )
    net.add_edge(2, 3, key="C", **{"EDGE_ID": "C", "LEN_": 100})

    node_list = [1, 2]
    edges = functions.get_edges_from_nodes(net, node_list)
    assert len(edges) == 1
    edges[0].key == "B"

    edges = functions.get_edges_from_nodes(net, [2, 1], with_direction_flag=True)
    assert len(edges) == 1
    assert edges[0].key == "B"
    assert edges[0].attributes[WITH_DIRECTION_FIELD] is False


def test_get_edges_from_nodes_non_unique():
    net = networkx.MultiGraph()

    net.add_edge(1, 2, key="A", **{"EDGE_ID": "A", "LEN_": 100})
    net.add_edge(1, 2, key="B", **{"EDGE_ID": "B", "LEN_": 100})

    node_list = [1, 2, 1]
    edges = functions.get_edges_from_nodes(net, node_list, return_unique=True)
    assert len(edges) == 1
    edges[0].key == "B"

    edges = functions.get_edges_from_nodes(net, node_list, return_unique=False)
    assert len(edges) == 2
    edges[0].key == "B"
    edges[1].key == "B"


def test_edges_to_graph():
    edges = []
    net = functions.edges_to_graph(edges)
    assert len(net.edges()) == 0

    edges = [Edge(0, 1, key=1, attributes={}), Edge(1, 2, key=2, attributes={})]
    net = functions.edges_to_graph(edges)
    assert len(net.edges()) == 2
    assert list(net.edges(keys=True)) == [(0, 1, 1), (1, 2, 2)]


def test_edges_to_graph_no_keys():
    tuples = [(0, 1), (1, 2)]
    net = functions.edges_to_graph(tuples)
    assert len(net.edges()) == 2
    # when no keys are set networkx uses 0 for all keys
    assert list(net.edges(keys=True)) == [(0, 1, 0), (1, 2, 0)]


def test_get_shortest_edge():
    edges = {
        "A": {"EDGE_ID": "A", "LEN_": 100, "NODEID_FROM": 1, "NODEID_TO": 2},
        "B": {"EDGE_ID": "B", "LEN_": 20, "NODEID_FROM": 2, "NODEID_TO": 3},
    }
    res = functions.get_shortest_edge(edges)
    assert res[0] == "B"


def test_get_shortest_edge_identical():
    """
    If two edges have identical lengths always
    return the higher value of the two keys
    """
    edges = {
        -1: {"EDGE_ID": 2, "LEN_": 100},
        1: {"EDGE_ID": 1, "LEN_": 100},
    }
    res = functions.get_shortest_edge(edges)
    assert res[0] == 1


def test_get_unique_ordered_list():
    lst = [2, 2, 1, 1, 4, 4, 1, 2, 3]
    res = functions.get_unique_ordered_list(lst)
    print(res)
    assert res == [2, 1, 4, 3]


def test_get_edges_from_node_pair():
    net = networkx.MultiGraph()

    net.add_edge(0, 1, key="A", val="foo")
    net.add_edge(0, 1, key="B", val="bar")

    tpls = functions.get_edges_from_node_pair(net, start_node=0, end_node=1)
    assert tpls == {"A": {"val": "foo"}, "B": {"val": "bar"}}


def test_get_path_length():
    edges = [Edge(0, 1, "A", {"LEN_": 5}), Edge(1, 2, "B", {"LEN_": 5})]
    assert functions.get_path_length(edges) == 10


def test_has_overlaps():

    edges = [
        Edge(0, 1, "A", {"EDGE_ID": 1}),
        Edge(1, 2, "B", {"EDGE_ID": 2}),
        Edge(2, 3, "C", {"EDGE_ID": 3}),
        Edge(2, 3, "C", {"EDGE_ID": 2}),
        Edge(2, 3, "C", {"EDGE_ID": 4}),
    ]
    assert functions.has_overlaps(edges) is True


def test_has_no_overlaps():

    edges = [
        Edge(0, 1, "A", {"EDGE_ID": 1}),
        Edge(1, 2, "B", {"EDGE_ID": 2}),
        Edge(2, 3, "C", {"EDGE_ID": 3}),
        Edge(2, 3, "C", {"EDGE_ID": 4}),
    ]
    assert functions.has_overlaps(edges) is False


def test_doctest():
    import doctest

    print(doctest.testmod(functions))


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    # test_get_edges_from_nodes()
    # test_edges_to_graph()
    # test_get_edges_from_nodes()
    # test_edges_to_graph()
    # test_get_shortest_edge()
    # test_get_shortest_edge_identical()
    # test_get_unique_ordered_list()
    # test_get_edges_from_node_pair()
    # test_to_edge()
    # test_to_edge_no_attributes()
    # test_edges_to_graph_no_keys()
    # test_get_edge_by_key_missing()
    # test_get_all_paths_from_nodes_with_direction()
    # test_get_path_length()
    # test_doctest()
    # test_get_edges_from_nodes_non_unique()
    test_has_no_overlaps()
    print("Done!")



from wayfarer import osmnx_compat
import wayfarer
from tests import networks
import logging
from shapely import wkb
from shapely.geometry import Point, LineString
import pytest


def test_to_osmnx_missing_geom():

    net = networks.simple_network()

    with pytest.raises(KeyError):
        osmnx_compat.to_osmnx(net)


def test_to_osmnx():

    net = networks.simple_network()

    assert len(net.edges()) == 4
    assert len(net.nodes()) == 5

    for e in wayfarer.to_edges(net.edges(keys=True, data=True)):
        e.attributes["Geometry4326"] = wkb.dumps(LineString([Point(0, 0), Point(0, 1)]))

    G = osmnx_compat.to_osmnx(net)

    assert len(G.edges()) == 8
    assert len(G.nodes()) == 5

    edges = list(wayfarer.to_edges(G.edges(keys=True, data=True)))

    keys = list(sorted(edges[0].attributes.keys()))

    # check all the edge keys for both wayfarer and osmnx are present
    assert keys == [
        "EDGE_ID",
        "Geometry4326",
        "LEN_",
        "NODEID_FROM",
        "NODEID_TO",
        "geometry",
        "highway",
        "length",
        "name",
        "oneway",
        "osmid",
        "reversed",
    ]


def test_doctest():
    import doctest

    print(doctest.testmod(osmnx_compat))


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    test_to_osmnx()
    print("Done!")



import logging
import wayfarer
from math import hypot
from wayfarer import (
    EDGE_ID_FIELD,
    LENGTH_FIELD,
    NODEID_FROM_FIELD,
    NODEID_TO_FIELD,
    GEOMETRY_FIELD,
)
from wayfarer import functions
import pickle
from typing import Iterable
from networkx import MultiGraph, MultiDiGraph


log = logging.getLogger("wayfarer")


class UniqueDict(dict):
    """
    A subclass of dict to ensure all keys are unique
    """

    def __setitem__(self, key, value):
        if key not in self:
            dict.__setitem__(self, key, value)
        else:
            raise KeyError("The key {} already exists. Keys must be unique".format(key))


def distance(
    p1: tuple[(int | float), (int | float)], p2: tuple[(int | float), (int | float)]
) -> float:
    """
    Return the Euclidean distance between two points.
    Any z-values associated with the points are ignored.

    Args:
        p1: The first point
        p2: The second point
    Returns:
        The distance between the two points

    >>> distance((0, 0), (10, 10))
    14.142135623730951
    """

    x1, y1 = p1[0], p1[1]
    x2, y2 = p2[0], p2[1]
    return hypot(x2 - x1, y2 - y1)


def save_network_to_file(net: (MultiGraph | MultiDiGraph), filename: str) -> None:
    """
    Save a network to a Python pickle file
    Note these cannot be shared between different versions of Python

    Args:
        net: The network
        filename: The filename to save the network as a pickle file

    """

    with open(filename, "wb") as f:
        pickle.dump(net, f, protocol=pickle.HIGHEST_PROTOCOL)


def load_network_from_file(
    filename: str,
) -> (MultiGraph | MultiDiGraph):
    """
    Return a network previously saved to a pickle file

    Args:
        filename: The filename to the pickle file containing the network
    Returns:
        The network

    """
    with open(filename, "rb") as f:
        return pickle.load(f)


def add_edge(net: (MultiGraph | MultiDiGraph), properties: dict) -> (str | int):
    """
    Add a new edge to a network based on a dict containing
    the required wayfarer fields

    Args:
        net: A network
        properties: A dictionary containing values for the edge
    Returns:
        The key of the edge

    >>> net = MultiGraph()
    >>> add_edge(net, {"EDGE_ID": 1, "NODEID_FROM": 1, "NODEID_TO": 1})
    1
    """

    key = properties[EDGE_ID_FIELD]
    start_node = properties[NODEID_FROM_FIELD]
    end_node = properties[NODEID_TO_FIELD]

    # key is used to identify different edges between nodes
    net.add_edge(start_node, end_node, key=key, **properties)

    # update the reverse lookup dictionary attached to the graph
    # this is used for fast edge_id lookups
    # the key is the dict key and the values are the nodes

    if "keys" in net.graph.keys():
        net.graph["keys"][key] = (start_node, end_node)

    return key


def create_graph(
    use_reverse_lookup: bool = True,
    graph_type: (MultiGraph | MultiDiGraph) = MultiGraph,
) -> (MultiGraph | MultiDiGraph):
    """
    Create a new networkx graph, with an optional dictionary to store unique keys
    for fast edge lookups

    >>> net = create_graph()

    Args:
        use_reverse_lookup: Create a dictionary for fast key lookup
        graph_type: The type of network
    Returns:
        The new network
    """
    if graph_type not in (MultiDiGraph, MultiGraph):
        raise ValueError(
            "The graph type {} unsupported. Only MultiGraph and MultiDiGraph are supported".format(
                graph_type.__name__
            )
        )

    graph_name = "Wayfarer Generated {} (version {})".format(
        graph_type.__name__, wayfarer.__version__
    )

    if use_reverse_lookup:
        reverse_lookup = UniqueDict()
        net = graph_type(name=graph_name, keys=reverse_lookup)
    else:
        net = graph_type(name=graph_name)
    return net


def load_network_from_records(
    recs: Iterable,
    use_reverse_lookup: bool = True,
    graph_type: (MultiGraph | MultiDiGraph) = MultiGraph,
    key_field: str = EDGE_ID_FIELD,
    length_field: str = LENGTH_FIELD,
    from_field: str = NODEID_FROM_FIELD,
    to_field: str = NODEID_TO_FIELD,
) -> (MultiGraph | MultiDiGraph):
    """
    Create a new networkX graph based on a list of dictionary objects
    containing the required wayfarer properties

    >>> recs = [{"EDGE_ID": 1, "NODEID_FROM": 1, "NODEID_TO": 2, "LEN_": 5, "PROP1": "A"}]
    >>> net = load_network_from_records(recs)
    >>> str(net)  # doctest: +ELLIPSIS
    "MultiGraph named 'Wayfarer Generated MultiGraph (version ...)' with 2 nodes and 1 edges"

    Args:
        recs: An iterable of dictionary objects The filename to the pickle file containing the network
        use_reverse_lookup: Create a dictionary as part of the graph that stores unique edge keys for
                            fast lookups
        graph_type: The type of network to create
    Returns:
        A new network

    """

    net = create_graph(use_reverse_lookup, graph_type)

    for r in recs:

        try:
            key = r[key_field]
            start_node = r[from_field]
            end_node = r[to_field]
            len_ = r[length_field]
        except KeyError:
            log.error("Available properties: {}".format(",".join(r.keys())))
            raise

        network_attributes = {
            EDGE_ID_FIELD: key,
            LENGTH_FIELD: len_,
            NODEID_FROM_FIELD: start_node,
            NODEID_TO_FIELD: end_node,
        }

        field_names = [key_field, length_field, from_field, to_field]
        all(map(r.pop, field_names))  # remove all keys already added
        network_attributes.update(r)  # add in any remaining keys

        add_edge(net, network_attributes)

    return net


def load_network_from_geometries(
    recs: Iterable,
    use_reverse_lookup: bool = True,
    graph_type: (MultiGraph | MultiDiGraph) = MultiGraph,
    skip_errors: bool = False,
    strip_properties: bool = False,
    keep_geometry: bool = False,
    key_field: str = EDGE_ID_FIELD,
    length_field: str = LENGTH_FIELD,
    rounding: int = 5,
) -> (MultiGraph | MultiDiGraph):
    """
    Create a new networkX graph using a list of recs of type ``__geo_interface__``
    This allows networks to be created using libraries such as Fiona.
    Any ``MultiLineString`` geometries in the network will be ignored, geometries should
    be converted to ``LineString`` prior to creating the network.

    >>> rec = {"geometry": {"type": "LineString", "coordinates": [(0, 0), (0, 1)]}, "properties": {"EDGE_ID": 1}}
    >>> net = load_network_from_geometries([rec])
    >>> str(net)  # doctest: +ELLIPSIS
    "MultiGraph named 'Wayfarer Generated MultiGraph (version ...)' with 2 nodes and 1 edges"

    Args:
        recs: An iterable of dictionary objects The filename to the pickle file containing the network
        use_reverse_lookup: Create a dictionary as part of the graph that stores unique edge keys for
                            fast lookups
        graph_type: The type of network to create
        skip_errors: Ignore any geometries that are not of type ``LineString``
        strip_properties: Reduce the size of the network file by only retaining the
                          properties required to run routing. Any other properties in the records
                          will be ignored
        key_field: A key field containing a unique Id for each feature must be provided
        length_field: The field in the geometry properties that contains a length. If not provided
                      then the length will be calculated from the geometry. It is assumed the geometry
                      is projected, as a simple planar distance will be calculated.
        rounding: If no node fields are used then nodes will be created based on the start and end points of
                  the input geometries. As node values must match exactly for edges to be connected rounding
                  to a fixed number of decimal places avoids unconnected edges to to tiny differences in floats
                  e.g. (-9.564484483347517, 52.421103202488965) and (-9.552925853749544, 52.41969110706263)
    Returns:
        A new network
    """

    net = create_graph(use_reverse_lookup, graph_type)
    error_count = 0

    for r in recs:

        # __geo__interface
        geom = r["geometry"]
        coords = geom["coordinates"]

        if geom["type"] != "LineString":
            if skip_errors is True:
                error_count += 1
                continue
            else:
                raise ValueError(
                    "The geometry type {} is not supported - only LineString can be used".format(
                        geom["type"]
                    )
                )

        properties = r["properties"]

        if length_field in properties:
            length = properties[length_field]
        else:
            length = sum([distance(*combo) for combo in functions.pairwise(coords)])

        if key_field == "id" and "id" in r:
            key = int(r["id"])
        else:
            try:
                key = int(properties[key_field])
            except KeyError:
                log.error(
                    "Available properties: {}".format(",".join(properties.keys()))
                )
                raise

        # if we simply take the coordinates then often these differ due to rounding issues as they
        # are floats e.g. (-9.564484483347517, 52.421103202488965) and (-9.552925853749544, 52.41969110706263)
        # instead convert the coordinates to unique strings, apply rounding, and strip any z-values

        start_node = "{}|{}".format(
            round(coords[0][0], rounding), round(coords[0][1], rounding)
        )
        end_node = "{}|{}".format(
            round(coords[-1][0], rounding), round(coords[-1][1], rounding)
        )

        network_attributes = {
            EDGE_ID_FIELD: key,
            LENGTH_FIELD: length,
            NODEID_FROM_FIELD: start_node,
            NODEID_TO_FIELD: end_node,
        }

        # only use fields required for routing
        if strip_properties is True:
            properties = network_attributes
        else:
            properties.update(network_attributes)

        if keep_geometry:
            properties[GEOMETRY_FIELD] = geom

        add_edge(net, properties)

    if error_count > 0:
        log.warning("{} MultiLineString features were ignored".format(error_count))

    return net



"""
This module handles splitting existing edges, and creating new nodes
to join the split edges
"""
import logging
import uuid
from collections import defaultdict, OrderedDict
from wayfarer import functions, linearref
import networkx
from wayfarer import (
    LENGTH_FIELD,
    OFFSET_FIELD,
    EDGE_ID_FIELD,
    SPLIT_FLAG,
    SPLIT_POINT_ID_FIELD,
    SPLIT_MEASURE_FIELD,
    NODEID_FROM_FIELD,
    NODEID_TO_FIELD,
    GEOMETRY_FIELD,
    Edge,
)

log = logging.getLogger("wayfarer")

SPLIT_KEY_SEPARATOR = "::"


def create_node_id(
    net: (networkx.MultiGraph | networkx.MultiDiGraph), point_id: (str | int)
):

    # if point_id in net.nodes():
    #    raise ValueError("The point ID {} already exists as a Node Id in the network".format(point_id))
    return point_id


def create_unique_join_node() -> str:
    """
    Create a unique GUID for the join node
    """

    return str(uuid.uuid4())


def group_measures_by_edge(input):

    res = OrderedDict()

    for edge_id, measure in input:
        if edge_id in res:
            res[edge_id].append(measure)
        else:
            res[edge_id] = [measure]
    return res


def split_with_points(
    net: (networkx.MultiGraph | networkx.MultiDiGraph),
    recs: list[dict],
    edge_id_field: str = EDGE_ID_FIELD,
    point_id_field: str = SPLIT_POINT_ID_FIELD,
    measure_field: str = SPLIT_MEASURE_FIELD,
):
    """
    Join a list of point records to a network, creating a new "join"
    edge

    recs is a list of dict objects that should have at a minimum the following
    fields [{"EDGE_ID": 1, "POINT_ID": 1, "MEASURE": 10.0}]
    Any additional fields will be added as properties to the new line
    """

    # first we'll group all points for each edge

    edge_splits = defaultdict(list)

    for r in recs:

        edge_id, point_id, measure = (
            r[edge_id_field],
            r[point_id_field],
            r[measure_field],
        )
        split_node = create_split_key(edge_id, measure)
        unique_edge_key = create_unique_join_node()
        from_node = create_node_id(net, point_id)

        # set the nodes as attributes so we can calculate the direction if required
        r[NODEID_FROM_FIELD] = from_node
        r[NODEID_TO_FIELD] = split_node

        # here we create a new edge joining the point to attach to the network
        # to a new node on the split line represented by the line id and measure along
        edge = Edge(
            start_node=from_node, end_node=split_node, key=unique_edge_key, attributes=r
        )

        functions.add_edge(net, **edge._asdict())  # unpack namedtuple to **kwargs
        edge_splits[edge_id].append(measure)

    for edge_id, measures in edge_splits.items():
        split_network_edge(net, edge_id, measures)


def get_split_attributes(
    original_attributes: dict,
    from_m: (float | int),
    to_m: (float | int),
    start_node: (str | int | None) = None,
    end_node: (str | int | None) = None,
) -> dict:
    """
    For a part of an edge that has been split
    update its attributes so the length is correct
    and any start and end nodes stored in the attributes table
    are correct

    >>> atts = {"OFFSET": 10, "LEN_": 100}
    >>> get_split_attributes(atts, 50, 70)
    {'OFFSET': 60, 'LEN_': 20, 'IS_SPLIT': True}
    """
    atts = original_attributes.copy()
    atts[LENGTH_FIELD] = to_m - from_m

    if OFFSET_FIELD in original_attributes:
        offset = original_attributes[
            OFFSET_FIELD
        ]  # when using a cumulative measure along a feature
    else:
        offset = 0

    atts[OFFSET_FIELD] = from_m + offset

    # update the nodes if provided and the fields are part of the list of attributes
    if start_node and end_node:
        if all(p in atts for p in [NODEID_FROM_FIELD, NODEID_TO_FIELD]):
            atts[NODEID_FROM_FIELD] = start_node
            atts[NODEID_TO_FIELD] = end_node

    # now add a flag to indicate this edge has been split
    atts[SPLIT_FLAG] = True

    # if geometry is being stored in the edge then create a new line
    # based on the offset
    if GEOMETRY_FIELD in atts:
        edge_id = atts[EDGE_ID_FIELD]
        ls = atts[GEOMETRY_FIELD]
        log.debug(f"Creating split geometry from {from_m} to {to_m} on {edge_id}")
        cut_line = linearref.create_line(ls, from_m, to_m)
        atts[GEOMETRY_FIELD] = cut_line

    return atts


def create_split_key(key: (str | int), m: (float | int)) -> str:
    """
    Function to ensure a standard format for split keys
    using the key of an edge and a measure
    E.g. "123829::154.2"
    The ``SPLIT_KEY_SEPARATOR`` can be used to find split keys
    in the network

    >>> SPLIT_KEY_SEPARATOR in "123829::154.2"
    True
    """
    return f"{key}{SPLIT_KEY_SEPARATOR}{m:g}"


def split_network_edge(
    net: (networkx.MultiGraph | networkx.MultiDiGraph),
    key: (str | int),
    measures: list[int | float],
) -> list[Edge]:
    """
    Split a network edge based on a list of measures.
    We can also split a split edge but this relies on knowing the key of the
    split edge and measures would need to reflect the length of the new edge
    rather than the original edge.

    Args:
        net (object): a networkx network
        key: the key of the edge to be split
        measures: a list of measures where the line should be split

    Returns:
        A list of the new edges created by splitting
    """

    new_edges = []

    original_edge = functions.get_edge_by_key(net, key, with_data=True)

    log.debug(f"Splitting {original_edge.key} with {len(measures)} points")

    net.remove_edge(original_edge.start_node, original_edge.end_node, original_edge.key)

    if "keys" in net.graph.keys():
        del net.graph["keys"][key]

    original_length = original_edge.attributes[LENGTH_FIELD]

    # measures = sorted(measures)

    for m in measures:
        if m >= original_length:
            raise ValueError(
                f"Split measure {m} is greater or equal to the length {original_length} of edge {key}"
            )
        if m == 0:
            raise ValueError("Split measure is 0 - no need to split!")

    prev_node = original_edge.start_node
    from_m = 0  # type: (int | float)

    for to_m in sorted(set(measures)):
        split_key = create_split_key(key, to_m)
        atts = get_split_attributes(
            original_edge.attributes, from_m, to_m, prev_node, split_key
        )
        new_edge = functions.add_edge(net, prev_node, split_key, split_key, atts)
        new_edges.append(new_edge)
        from_m, prev_node = to_m, split_key

    # add final part
    atts = get_split_attributes(
        original_edge.attributes,
        from_m,
        original_length,
        prev_node,
        original_edge.end_node,
    )

    split_key = create_split_key(key, original_length)
    new_edge = functions.add_edge(
        net, prev_node, original_edge.end_node, split_key, atts
    )
    new_edges.append(new_edge)

    return new_edges


def get_measure_for_point(line, pt):
    snapped_input_point, dist = linearref.get_nearest_vertex(pt, line)
    log.debug(f"Input point {dist:.5f} from line")
    measure = linearref.get_measure_on_line(line, snapped_input_point)
    return measure


def get_split_node_for_measure(network_edge, length, measure):

    if abs(length - measure) < 0.001:
        # don't split at the end of the line
        node_id = network_edge.end_node
    elif abs(measure - 0) < 0.001:
        # don't split at the start of the line
        node_id = network_edge.start_node
    else:
        # a new split node is created
        node_id = create_split_key(network_edge.key, measure)

    return node_id



"""
Helper module to convert between wayfarer and osmnx networks
"""
import wayfarer
import networkx as nx
from shapely import wkb, LineString
import osmnx as ox
import logging
import datetime as dt


from wayfarer import (
    LENGTH_FIELD,
    NODEID_FROM_FIELD,
    NODEID_TO_FIELD,
)


log = logging.getLogger("wayfarer")


def to_osmnx(
    net: (nx.MultiGraph | nx.MultiDiGraph),
    geometry_field: str = "Geometry4326",
    crs: str = "epsg:4326",
) -> nx.MultiDiGraph:
    """
    Convert a wayfarer network into a osmnx network
    default_crs in osmnx is in settings.default_crs and is epsg:4326
    """

    # for timestamp see code from https://github.com/gboeing/osmnx/blob/3822ed659f1cc9f426a990c02dd8ca6b3f4d56d7/osmnx/utils.py#L50
    template = "{:%Y-%m-%d %H:%M:%S}"
    ts = template.format(dt.datetime.now())

    metadata = {
        "created_date": ts,
        "created_with": f"OSMnx {ox.__version__}",
        "crs": crs,
    }

    G = nx.MultiDiGraph(**metadata)  # wayfarer typically uses MultiGraph networks

    for tpl in net.edges(data=True, keys=True):

        edge = wayfarer.to_edge(tpl)

        try:
            edge_geometry = edge.attributes[geometry_field]
        except KeyError:
            keys = list(edge.attributes.keys())
            log.error(
                f"The attribute '{geometry_field}' was not found in the edge attributes: {keys}"
            )
            raise

        if not type(edge_geometry) is LineString:
            edge_geometry = wkb.loads(edge_geometry)

        G.add_node(
            edge.start_node, x=edge_geometry.coords[0][0], y=edge_geometry.coords[0][1]
        )
        G.add_node(
            edge.end_node, x=edge_geometry.coords[-1][0], y=edge_geometry.coords[-1][1]
        )

        properties = {
            "osmid": edge.key,
            "name": "test",
            "highway": "highway",
            "oneway": False,
            "reversed": False,
            "length": edge.attributes[LENGTH_FIELD],
            "geometry": edge_geometry,
        }

        # merge with wayfarer fields
        edge.attributes.update(properties)
        G.add_edge(edge.start_node, edge.end_node, key=edge.key, **edge.attributes)

        # now add the reversed edge
        reversed_properties = edge.attributes.copy()
        reversed_properties.update(
            {
                "reversed": True,
                NODEID_FROM_FIELD: edge.end_node,
                NODEID_TO_FIELD: edge.start_node,
            }
        )
        key = edge.key * -1

        G.add_edge(edge.end_node, edge.start_node, key=key, **reversed_properties)

    return G



from wayfarer import loops, Edge
import logging
from tests import networks


def test_get_root_node():

    outdeg = [(0, 2), (1, 3), (2, 2)]
    root_node = loops.get_root_node(outdeg)
    assert root_node == 1

    outdeg = [(0, 2), (1, 2), (2, 2)]
    root_node = loops.get_root_node(outdeg)
    assert root_node == 0


def test_get_unique_lists():

    all_lists = [[1, 2, 3], [1, 3, 2]]
    unique_lists = loops.get_unique_lists(all_lists)
    assert unique_lists == [[1, 2, 3]]

    all_lists = [[1, 2, 3], [1, 3, 2, 4], [1, 3, 2]]
    unique_lists = loops.get_unique_lists(all_lists)
    assert unique_lists == [[1, 2, 3], [1, 3, 2, 4]]

    all_lists = [[1, 2], [2, 1], [1, 2]]
    unique_lists = loops.get_unique_lists(all_lists)
    assert unique_lists == [[1, 2]]


def test_get_loop_nodes():

    edges = [Edge(0, 1, "A", {}), Edge(1, 2, "B", {}), Edge(2, 0, "C", {})]
    loop_nodes = loops.get_loop_nodes(edges)
    assert loop_nodes == {0: [0, 1, 2]}


def test_is_loop():

    assert loops.is_loop(networks.circle_network()) is True


def test_is_not_loop():

    assert loops.is_loop(networks.triple_loop_network()) is False


def test_doctest():
    import doctest

    print(doctest.testmod(loops))


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    # test_get_unique_lists()
    # test_get_root_node()
    # test_get_loop_nodes()
    # test_is_loop()
    test_is_not_loop()
    print("Done!")



import re
from setuptools import setup
from io import open

(__version__,) = re.findall('__version__ = "(.*)"', open("wayfarer/__init__.py").read())


def readme():
    with open("README.rst", "r", encoding="utf-8") as f:
        return f.read()


setup(
    name="wayfarer",
    version=__version__,
    description="A library to add spatial functions to networkx",
    long_description=readme(),
    classifiers=[
        # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",  # minimum required due to use of annotations
        "Intended Audience :: Developers",
    ],
    url="https://github.com/compassinformatics/wayfarer/",
    author="Seth Girvin",
    author_email="sgirvin@compass.ie",
    license="MIT",
    package_data={"wayfarer": ["py.typed"]},
    packages=["wayfarer"],
    install_requires=["networkx>=3.0", "geojson"],
    zip_safe=False,
)
